from math import ceil

class DimSymbol():

    decls = {}
    __slots__ = 'name', 'fullname'

    def __init__(self, name:str, fullname:str =None):
        self.name = name
        self.fullname = fullname
        DimSymbol._declare(self)


    @property
    def longname(self):
        res = '' if self.fullname is None else self.fullname
        res += f'({self.name})'
        return res
    

    @staticmethod
    def _declare(d):
        if d.name in DimSymbol.decls:
            raise ValueError(f'Dimension {d.name} already declared')
        DimSymbol.decls[d.name] = d

    @staticmethod
    def lookup(name: str, hard=True):
        if name not in DimSymbol.decls:
            if hard:
                assert False, (f'Dim {name} not declared')
            else: return None
        else:
            return DimSymbol.decls[name]

    def __repr__(self):
        return self.name

    '''
    @staticmethod
    def eval_name(e):
        #evalute a shape expression, in context of declared names
        sub_map = [(e, dv.longname) for e, dv in Dim.decls.items()]
        return str(e.subs(sub_map))
    '''


class ContinuousDim(DimSymbol):
    __slots__ = 'span', '_size'
    def __init__(self, name: str, span: slice, fullname: str=None):
        super().__init__(name=name, fullname=fullname)
        self.span = span

        size = (span.stop - span.start)
        if span.step is not None:
            size = ceil( (1.* size)/span.step) # (1,8,2) -> 8 - 1 = 7. ceil(7 / 2)
        self._size = size


    def size(self):
        return self._size

    def __repr__(self):
        return super().__repr__()

class SingletonDim(DimSymbol):
    __slots__ = 'valtype'
    def __init__(self, name: str, valtype: str, fullname: str=None):
        super().__init__(name=name, fullname=fullname)
        self.valtype = valtype


class CategoricalDim(DimSymbol):
    __slots__ = 'categories', '_num_categories'

    def __init__(self, name, categories, fullname: str=None):
        super().__init__(name=name, fullname=fullname)
        self.categories = [SingletonDim(name=cat[0], valtype=cat[1]) for cat in categories]
        self._num_categories = len(self.categories)
    def num_categories(self):
        return self._num_categories

from typing import Union

Dim = Union[ContinuousDim, CategoricalDim]



from .aos import get_or_decl_dim, AOShape, AOop
from .utils import apply_match_t2a

class InferAOS():
    def __init__(self, max_list_items=10, simplify=True):
        self.max_list_items = max_list_items
        self.simplify = simplify

        self.type2action = {
            dict: self.for_dict,
            (list, tuple): self.for_list_tuple,
            (str, int, float, bool): self.for_scalar,
        }

    def for_dict(self, obj, **kwargs):
        res = []
        for k, v in obj.items():
            k_ = get_or_decl_dim(k)
            v_ = self.infer(v)
            kv = AOShape.build_from(AOop.AND, [k_, v_])
            res.append(kv)

        res = AOShape.build_from(AOop.OR, res)
        return res

    def for_list_tuple(self, obj, **kwargs):
        max_list_items = self.max_list_items

        if len(obj) > max_list_items:
            print (f'Warning: only observing {max_list_items}/{len(obj)} items in list!')
        args = [self.infer(v) for v in obj[:max_list_items]]
        res = AOShape.build_from(AOop.OR, args)

        if self.simplify:
            args_ = list(set(res.args))
            if len(args_) == 1:
                res = AOShape.build_from(AOop.SEQUENCE, args_)
            if len(args_) < len(res.args):
                print ('Suggestion: list aos can be potentially compressed')

        return res

    def for_scalar(self, obj, **kwargs):
        return get_or_decl_dim(type(obj).__name__)

    def default_func(self, obj, **kwargs):
        raise NotImplementedError(f'InferAOS : type = {type(obj), type(obj).__name__}')

    def infer(self, obj):
        return apply_match_t2a(self.type2action, obj, default=self.default_func)

def infer_aos (x, max_list_items=10, simplify=True):
    res = InferAOS(max_list_items, simplify).infer(x)

    return res























from .dim import Dim, DimSymbol, CategoricalDim, ContinuousDim
from .common import Config
from enum import Enum
from typing import NamedTuple, List
from pprint import pformat

class AOop(Enum):
    AND: int = 0
    OR: int = 1
    COLON: int = 2
    COMMA: int = 3
    SEQUENCE: int = 4

    @staticmethod
    def to_str(op):
        if op == AOop.AND:
            return '&'
        elif op == AOop.OR:
            return '|'
        elif op == AOop.COLON:
            return ':'
        elif op == AOop.SEQUENCE:
            return '*'
        else:
            assert False

def is_dim_instance(shape):
    #check Union
    #if shape.__origin__ is Union:
    return shape in Dim.__args__


class AOConst(NamedTuple):
    UNDERSCORE: str = '_'
    ELLIPSIS: str = '...'

    @staticmethod
    def is_underscore(s: str):
        return s == AOConst().UNDERSCORE

    @staticmethod
    def is_ellipsis(s: str):
        return s == AOConst().ELLIPSIS




'''
Shape = Dim | AndDim of Shape*| OrDim of Shape*
Shape = Dim | Op of op, Shape*
'''

'''
AOShape :=
    ContinuousDim of (name: str, span: Slice)
    | SingletonDim of (name: str, valtype: str, logicaltype: str)
    | CategoricalDim of (name: str, categories: List[SingletonDim])
    | Op of (opt_name, op, List[AOShape])

    # defn without recursion:
    | AOShape of (optname, sexpr) #sexpr is a separate recursive type

'''

class AOShape(NamedTuple):
    op: int = None
    args: 'List[AOShape]' = None
    name: str = None

    dim: Dim = None


    optext2op = {
    '&': AOop.AND,
    '|': AOop.OR,
    ':': AOop.COLON,
    ',': AOop.COMMA,
    '*': AOop.SEQUENCE
    }

    def get_dim_name(self):
        assert self.dim is not None
        return self.dim.name.strip()


    @staticmethod
    def build_from(op, args):
        #print (f'build_from : {op}, {args}')

        if isinstance(op, str):
            op = AOShape.optext2op[op]
        for arg in args:
            assert isinstance(arg, AOShape), f'{arg}'

        if len(args) > 2 and op is AOop.AND: #convert n-ary to a binary tree
            args = args[::-1] #reverse list
            b, a = args[:2] # order of last 2 elements is reversed
            res = AOShape(op=op, args=(a,b))
            for arg in args[2:]:
                res = AOShape(op=op, args=[arg, res])
        else:
            res = AOShape(op=op, args=args)
        #print (f'build_from ret: {AOop.to_str(res.op)}')
        return res

    def make_shape(self, shape):
        assert isinstance(shape, AOShape)

        if is_dim_instance(shape):
            shape = AOShape(dim=shape)
        else:
            assert isinstance(shape, AOShape)
        return shape

    def andop(self, shape, name=None):

        shape = self.make_shape(shape)
        ret = AOShape(op=AOop.AND, args=[self,shape], name=name)
        return ret

    def __and__(self, shape):
        return self.andop(shape)
    def __rand__(self, shape):
        return self.andop(shape)

    def orop(self, shape, name=None):
        shape = self.make_shape(shape)
        ret = AOShape(op=AOop.OR, args=[self,shape], name=name)
        return ret

    def __or__(self, shape):
        return self.orop(shape)
    def __ror__(self, shape):
        return self.orop(shape)

    def __key(self):
        if self.dim is not None:
            return self.dim
        else:
            res = (self.op, self.name) + tuple(self.args)
        return res

    def __hash__(self):
        return hash(self.__key())

    def __repr_obj__(self, depth=0):
        res = {}

        if self.op is None:
            assert self.dim is not None
            res = self.dim.__repr__()
        else:
            op = AOop.to_str(self.op)
            args = [a.__repr_obj__(depth+1) for a in self.args]

            if self.op is AOop.OR:
                res = args
            else:
                res = (op, args)
                if self.op is AOop.AND:
                    arg0 = args[0]
                    rest_args = args[1:] if len(args) > 2 else args[-1] 
                    res = (arg0, rest_args)
                elif self.op is AOop.SEQUENCE:
                    res = {}
                    assert len(args) == 1
                    res[op] = args[0]

        return res

    def __repr__(self):
        if self.op is None:
            assert self.dim is not None
            return self.dim.__repr__()
        else:
            #res = [f'{a} {op}' for i, a in args]
            if not Config.pprint_treelike:
                op = AOop.to_str(self.op)
                #print (f'rep: op = {self.op}, {op}')
                args = [a.__repr__() for a in self.args]

                res = []
                # [x for x in chain.from_iterable(zip_longest(l1, l2)) if x is not None]
                for i, a in enumerate(args):
                    if i != len(args) - 1:
                        res.append(f'{a} {op}')
                    else:
                        res.append(f'{a}')
                #print (f'repr: {res}')
                if op == '*': res.append('*')

                res = '(' + ' '.join(res) + ')'
            
            else:
                res = self.__repr_obj__()
                res = pformat(res, indent=4)

            return res


    def __str__(self):
        return str(self.__repr__())



def decl_dim (name, dimtype, values):
    if dimtype == 'categorical':
        #TODO: convert values to categories
        dim = CategoricalDim(name=name, categories=values)
    elif dimtype == 'continuous':
        #TODO: convert values to span(slice)
        dim = ContinuousDim(name=name, span=values)
    return dim


def get_dims(names):
    '''
    names: 'b c h w', separated by spaces
    '''
    names = names.strip().split(' ')
    res = [AOShape(dim=DimSymbol.lookup(name)) for name in names]
    if len(names) == 1: return res[0]
    else: return res

def get_or_decl_dim(name):
    dim = DimSymbol.lookup(name, hard=False)
    if dim is None:
        dim = decl_dim(name=name, dimtype='categorical', values=[])
    return AOShape(dim=dim)

def get_or_decl_dims(names):
    names = names.strip().split(' ')
    res = [get_or_decl_dim(name) for name in names]
    if len(names) == 1: return res[0]
    else: return res


def show_dims(names):
    names = names.strip().split(' ')
    res = [str(DimSymbol.lookup(name)) for name in names]
    return '\n'.join(res)
















import sys
from itertools import chain

from .parser import parse_aos
from .dim import DimSymbol, ContinuousDim, CategoricalDim, Dim
from .aos import AOShape, AOop, AOConst
from .type_matchers import OrList, AndMatcher
from .utils import apply_match, pair_items

from .common import Config, AndTuple
DEBUG = Config.DEBUG

def as_or_list(obj):
    return apply_match(OrList, obj)

def get_or_match_args (obj, shape):
    res, err = None, None
    shape_args = shape.args

    obj_args, err = as_or_list(obj)
    if err is not None: return res, err

    if len(obj_args) != len(shape_args):
        err = f'shape mismatch: {obj_args}, {shape_args}'
        return res, err
    res = zip(obj_args, shape_args)
    return res, err


def get_and_match_args(obj, shape, exact=False, sample=False):
    shape_args = shape.args
    return apply_match(AndMatcher, obj, shape_args, exact=exact, sample=sample)

def get_sequence_match_args(obj, shape, exact=True, sample=False):

    #obj : list(x1, x2, ..), shape = (c|d)*
    # match(x1, 'c|d')
    res, err = None, None
    assert len(shape.args) == 1
    shape_arg = shape.args[0]

    if isinstance(obj, list):
        res = [(o, shape_arg) for o in obj]
    elif isinstance(obj, dict):
        res = [(AndTuple(item), shape_arg) for item in obj.items()]
    else:
        err = f'shape mismatch with sequence: {type(obj)}'

    return res, err

def get_match_args (op, obj, shape):
    '''
    return a list of (obj, shape) pairs to match
    '''
    if DEBUG: print(f'get_match_args: {op}, {obj}, {shape}')
    if op == AOop.OR:
        res = get_or_match_args(obj, shape)
    elif op == AOop.AND:
        res = get_and_match_args(obj, shape)
    elif op == AOop.SEQUENCE:
        res = get_sequence_match_args(obj, shape)
    else:
        raise NotImplementedError(f'op: {op}, shape: {shape}')

    if DEBUG: print (f'< get_match_args: {res}')
    return res

def is_scalar(obj):
    return isinstance(obj, (str, int, float, bool))

def is_dim(obj):
    if isinstance(obj, (ContinuousDim, CategoricalDim)): return True
    if isinstance(obj, str):
        return (DimSymbol.lookup(obj, hard=False) is not None)
    return False

def is_dim_or_scalar(obj):
    return is_scalar(obj) or is_dim(obj)

def match_obj_const(obj, const):

    if isinstance(obj, int) or (isinstance(obj, str) and obj.isdigit()):
        return int(obj) == const

    dim = DimSymbol.lookup(obj, hard=False)
    if dim is None: return False

    if isinstance(dim, ContinuousDim):
        return dim.size() == const
    return False


def match_dim(obj, dim_name):
    #print (f'matching {obj}, {dim_name}')
    err = None
    if DEBUG: print (f'match_dim: dname = {dim_name}, {type(dim_name)}')
    if AOConst.is_ellipsis(dim_name):  #match anything
        pass
    else:
        test = False
        if dim_name.isdigit():
            test = match_obj_const(obj, int(dim_name))
        else:
            test = (obj is sys.intern(dim_name)) \
                or sys.intern(type(obj).__name__) is sys.intern(dim_name)\
                or (is_dim_or_scalar(obj) and (AOConst.is_underscore(dim_name))\
                )

        err = None if test\
            else f'** Leaf mismatch: obj=({obj}), dim=({dim_name}) **'

    return err

def match_obj_shape(obj, shape) -> 'err':
    err = None
    if DEBUG: print (f'match_obj_shape: {obj}, {shape}')

    if shape.op is None:
        dname = shape.dim.name.strip()
        err = match_dim(obj, dname)

    else:

        match_pairs, err = get_match_args(shape.op, obj, shape)
        if err is not None: return err

        for pair in match_pairs:
            err = match_obj_shape(pair[0], pair[1])
            if err is not None:
                return err

    return err

def instanceof(arr, shape):
    if isinstance(shape, str):
        shape = parse_aos(shape)
    
    if DEBUG:
        print (f'\nchecking {arr} against {shape}')

    err = match_obj_shape(arr, shape)
    if err: 
        print (err)
        return False, err
    else:
        return True, None








from types import SimpleNamespace
II = isinstance


#from typing import NewType
#AndTuple = NewType('AndTuple', tuple)
class AndTuple(tuple):
    def __repr__(self):
        s = super().__repr__()
        return f'`{s}`'

def simplify_and2dict(obj):
    if isinstance(obj, AndTuple):
        k, v = obj
        #assert isinstance(k, str)
        res = {k: simplify_and2dict(v)}
    elif isinstance(obj, dict):
        res = {k: simplify_and2dict(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        ty = type(obj)
        res = ty([simplify_and2dict(x) for x in obj])
    elif isinstance(obj, (int, float, str, bool)):
        res = obj

    else:
        raise ValueError(f'not handled type {type(obj)}')

    return res


class OrVals(list):
    # for storing a list of disjoint contexts
    def __repr__(self):
        s = super().__repr__()
        return f'|{s}|'

    @staticmethod
    def merge (ovlist: 'List[OrVals]'):
        lens = [len(ov) for ov in ovlist]
        assert lens.count(lens[0]) == len(lens), 'all OrVals not of same length'

        pairs = list(zip(*ovlist))
        #print (f'merge: {ovlist}, \n{list(pairs)}')
        res = ([OrVals(p).simplify() for p in pairs])

        res = OrVals(res) #BUG? do we need additional OrVals here ot List
        return res

    def simplify(self):
        # list of AndTuple -> dict
        for x in self:
            if not II(x, AndTuple): return self
            #z = dict(x)
        res = dict(self)
        return res
    

config = dict(
    pprint_treelike=False, 
    DEBUG=False
)
Config = SimpleNamespace(**config)


from aos.parser import parse_aos
from aos.bind_matchers import bind_obj_shape
from aos.eval_aos import eval_aos_in_context
from aos.common import Config, simplify_and2dict
DEBUG = Config.DEBUG

def parse_tfm(tfm: str):
    lhs, rhs = [x.strip() for x in tfm.split('->')]
    lhs = parse_aos(lhs)
    rhs = parse_aos(rhs)
    if DEBUG: print (lhs, rhs)
    return lhs, rhs

def do_tfm(obj, tfm: str):
    if DEBUG: print (f'\ntfm -> {tfm}, {obj}')
    lhs, rhs = parse_tfm(tfm)
    context, err =  bind_obj_shape(obj, lhs)
    if err is not None:
        raise ValueError(err)
    if DEBUG: print ('bindings:', context)
    #assert False

    out, err = eval_aos_in_context(rhs, context)
    #print('simplify_and2dict')
    out = simplify_and2dict(out)
    return out


from itertools import chain

from .common import AndTuple
from .utils import pair_items

class OrList():
    '''
    convert an or-like object into a list of objects
    '''
    def from_AndTuple(arr):
        err = f'mismatch: cannot convert AndTuple to Or-shape {arr}'
        return None, err
    #def from_list(arr):
    #    return arr, None
    #def from_tuple(arr):
    #    return list(arr), None
    def from_dict(arr):
        #convert into a list of AND types (ANDtuple)
        items = [AndTuple(x) for x in arr.items()]
        return items, None
    def default_func(arr):
        err = f'mismatch: expected list or dict: unknown arr type {type(arr)}'
        return None, err

    type2action = {
        list: lambda arr: (arr, None),
        AndTuple: from_AndTuple,
        tuple: lambda arr: (list(arr), None),
        dict: from_dict
    }

    @classmethod
    def add_action(cls, klass, func):
        assert klass not in cls.type2action
        cls.type2action[klass] = func




class AndMatcher():
    def for_dict(obj, shape_args, **kwargs):
        res, err = None, None
        if not kwargs['exact']: 
            res, err = pair_items(obj.items(), shape_args, sample=kwargs['sample'])
        else:
            err = f'shape mismatch: dict cannot match with AND: \n{dict}\n{shape_args}'
        return res, err

    def for_dataframe(obj, shape_args, **kwargs):
        res, err = None, None
        n = len(obj)
        if len(shape_args) != 2:
            err = f'mismatch: shape {shape_args} with pandas DataFrame'
        else:
            res = zip([n]+ [list(obj)], shape_args) 
        return res, err

    def for_AndTuple(obj, shape_args, **kwargs):
        res, err = pair_items([obj], shape_args, sample=False) 
        return res, err

    def for_ndarray(arr, shape_args, **kwargs):
        res, err = None, None
        shape = arr.shape
        #print(f'{arr.shape}, {shape_args}')
        if len(shape) != len(shape_args):
            err = f'mismatch: shape {shape} against args {shape_args}'
        else:
            res = zip(shape, shape_args)
        return res, err
    
    def for_torch_tensor(arr, shape_args, **kwargs):
        res, err = None, None
        shape = arr.size()
        #print(f'{arr.shape}, {shape_args}')
        if len(shape) != len(shape_args):
            err = f'mismatch: shape {shape} against args {shape_args}'
        else:
            res = zip(shape, shape_args)
        return res, err


    def default_func(obj, shape_args, **kwargs):
        raise NotImplementedError(f'AndMatcher : type = {type(obj), type(obj).__name__}')

    type2action = {
        dict: for_dict,
        AndTuple: for_AndTuple,
        'DataFrame': for_dataframe,
        'ndarray': for_ndarray,
        'torch.Tensor': for_torch_tensor
    }





from typing import Callable
from itertools import chain

def unroll_type2actions(t2a):
    res = {}
    for k, a in t2a.items():
        if isinstance(k, tuple):
            for i in k:
                assert i not in res
                res[i] = a
        else:
            res[k] = a
    return res

def apply_match_t2a(t2a, obj, *args, default_func=None, **kwargs):
    '''
        t2a: type2action dictionary for the various 'obj' types
    '''
    t2a = unroll_type2actions(t2a)
    tobj = type(obj)
    #print(f'apply_match: {tobj}, {tobj.__name__}')
    if tobj in t2a:
        tkey = tobj
    elif tobj.__name__ in t2a:
        tkey = tobj.__name__
    elif (f'{tobj.__module__}.{tobj.__name__}') in t2a:
        tkey = f'{tobj.__module__}.{tobj.__name__}'
    else:
        print (f'apply_match: not found {tobj} key')
        assert default_func, 'no default type handler'
        return default_func(obj, *args, **kwargs)     
    
    func = t2a[tkey]
    if isinstance(func, staticmethod):
        func = func.__func__ 
    else:
        assert isinstance(func, Callable), f'apply_match: type = {type(func)}'

    return func(obj, *args, **kwargs)

def apply_match(klass, obj, *args, **kwargs):
    #print(f'from_obj: {obj}, {type(obj)}')
    t2a = klass.type2action
    default_func = getattr(klass, 'default_func', None)
    return apply_match_t2a(t2a, obj, *args, default_func=default_func, **kwargs)


def pair_items(items: 'iterator', shape_args: 'list', sample=False):
    '''
    pair some or all items from <items> with <shape_args>
    '''
    res, err = None, None

    if sample:
        item = next(iter(items))
        if len(item) != len(shape_args):
            return None, f'shape mismatch: {item}, {shape_args}'
        res = zip(item, shape_args)
    else:
        pairs = []
        for item in items:
            if len(item) != len(shape_args):
                return None, f'shape mismatch: {item}, {shape_args}'
            else:
                pairs.append(zip(item, shape_args))
        res = chain.from_iterable(pairs)

    return res, err
